import argparse
from datetime import datetime
import os
from queue import Empty
import signal
import logging
from sys import stderr
from threading import Event
from multiprocessing import Process, Queue
from time import time
import isys6030
from isys6030.version import __version__


def configure_logging():
    logging.basicConfig(format='%(asctime)s %(levelname)-10s%(message)s', level=logging.INFO)


def usage():
    parser = argparse.ArgumentParser(description="iSYS-6030 CLI")
    parser.add_argument('--port', type=str, help='Sensor port')
    parser.add_argument('--baud-rate', type=int, default=115200, help='Sensor baud rate')
    parser.add_argument('--sensor-address', type=int, default=100, help='Sensor address')
    parser.add_argument('--minimum-range', type=float, help='Minimum range')
    parser.add_argument('--maximum-range', type=float, help='Maximum range')
    parser.add_argument('--minimum-signal', type=float, help='Minimum signal')
    parser.add_argument('--spectrum', action='store_const', const=True, help='Fetch complete spectrum')
    parser.add_argument('--threshold', type=float, help='Signal threshold')
    parser.add_argument('--filter-type', type=int, help='''\
        Type of signal target filter (0 = highest amplitude, 1 = mean, 2 = median, 3 = min (closest), 4 = max (furthest))''')
    parser.add_argument('--filter-signal', type=int, help='Signal for singel target filter (0 = off, 2 = range radial)')
    parser.add_argument('--save', action='store_const', const=True, help='Save settings to sensor flash')
    parser.add_argument('--stdout', action='store_const', const=True, help='Dump range and signal data to standard output')
    parser.add_argument('--path', type=str, help='Save range and signal data to files on this path')
    parser.add_argument('--verbose', action='store_const', const=True, help='More verbose output')
    parser.add_argument('--version', action='store_const', const=True, help='Show version and exit')
    return parser


def format_spectrum(timestamp, spectrum):
    return f"{timestamp:.3f} {' '.join(map(str, spectrum))}"

def format_value(timestamp, rangevalue, signal_strength):
    return f'{timestamp:.3f} {rangevalue:.4f} {signal_strength:.2f}'

def write_files(queue: Queue, data_path, serial_number):
    configure_logging()

    current_file_time = None
    data_file = None
    while True:
        try:
            item = queue.get(timeout=1.0)
        except TimeoutError:
            continue
        except Empty:
            continue

        if item is None:
            return

        try:
            current_datetime = datetime.fromtimestamp(item['timestamp'])
            file_time = datetime(current_datetime.year, current_datetime.month, current_datetime.day, current_datetime.hour)

            if file_time != current_file_time:
                if data_file:
                    data_file.close()
                file_start_timestamp = file_time.timestamp()
                output_dir = os.path.join(data_path, f'{serial_number:04d}',
                                          file_time.strftime(f'%Y{os.path.sep}%m{os.path.sep}%d'))
                if not os.path.exists(output_dir):
                    os.makedirs(output_dir)

                filename = f"ISYS_{serial_number:04d}_{file_time.strftime('%Y%m%d_%H%M%S')}.dat"
                logging.info('Writing to file %s', filename)
                data_file = open(os.path.join(output_dir, filename), 'a', encoding='ascii')
                current_file_time = file_time

            if 'spectrum' in item:
                data_file.write(format_spectrum(item['timestamp'] - file_start_timestamp, item['spectrum']) + "\n")
            else:
                data_file.write(format_value(item['timestamp'] - file_start_timestamp, item['range'], item['signal_strength']) + "\n")
        except Exception:  # pylint: disable=broad-except
            logging.exception('Error writing data')


def main():
    configure_logging()
    terminated = Event()
    signal.signal(signal.SIGINT, lambda *_: terminated.set())
    signal.signal(signal.SIGTERM, lambda *_: terminated.set())
    file_output_queue = None
    file_output_process = None

    parser = usage()
    config = parser.parse_args()
    if config.version:
        print(f'isys6030-cli version {__version__}')
        return

    if not config.port:
        parser.print_usage()
        stderr.write('error: the following argument is required: --port\n')
        return

    with isys6030.Sensor(config.port, baudrate=config.baud_rate, sensor_address=config.sensor_address) as sensor:
        device_name, serial_number = sensor.get_device_name()
        if config.minimum_range is not None:
            sensor.set_minimum_range(config.minimum_range)
        if config.maximum_range is not None:
            sensor.set_maximum_range(config.maximum_range)
        if config.minimum_signal is not None:
            sensor.set_minimum_signal(config.minimum_signal)
        if config.filter_type is not None:
            sensor.set_filter_type(config.filter_type)
        if config.filter_signal is not None:
            sensor.set_filter_signal(config.filter_signal)
        if config.threshold is not None:
            sensor.set_threshold(config.threshold)
        if config.save:
            sensor.save_settings()

        if config.verbose:
            minimum_range = sensor.get_minimum_range()
            maximum_range = sensor.get_maximum_range()
            minimum_signal = sensor.get_minimum_signal()
            filter_type = sensor.get_filter_type()
            filter_signal = sensor.get_filter_signal()
            threshold = sensor.get_threshold()
            firmware_version = sensor.get_firmware_version()
            hardware_version = sensor.get_hardware_version()
            product_code = sensor.get_product_code()
            bootloader_version = sensor.get_bootloader_version()

            if filter_type == 0:
                filter_type_str = 'highest amplitude'
            elif filter_type == 1:
                filter_type_str = 'mean'
            elif filter_type == 2:
                filter_type_str = 'median'
            elif filter_type == 3:
                filter_type_str = 'min (closest)'
            elif filter_type == 4:
                filter_type_str = 'max (furthest)'
            else:
                filter_type_str = 'unknown'

            if filter_signal == 0:
                filter_signal_str = 'off'
            elif filter_signal == 2:
                filter_signal_str = 'range radial'
            else:
                filter_signal_str = 'unknown'

            print(f'Device name:        {device_name}')
            print(f'Serial number:      {serial_number}')
            print(f'Product code:       {product_code}')
            print(f'Hardware version:   {hardware_version}')
            print(f'Bootloader version: {bootloader_version}')
            print(f'Firmware version:   {firmware_version}')
            print(f'Filter type:        {filter_type_str}')
            print(f'Filter signal:      {filter_signal_str}')
            print(f'Minimum range:      {minimum_range:5.1f} m')
            print(f'Maximum range:      {maximum_range:5.1f} m')
            print(f'Minimum signal:     {minimum_signal:5.1f} dB')
            print(f'Threshold:          {threshold:5.1f} dB')

        if config.path:
            file_output_queue = Queue()
            file_output_process = Process(target=write_files, args=(file_output_queue, config.path, serial_number))
            file_output_process.start()

        start_time = time()
        sensor.start_acquisition()
        sensor.set_mode(0)

        while not terminated.is_set():
            if config.spectrum:
                try:
                    spectrum, _, _ = sensor.get_range_spectrum()
                    timestamp = time()
                except TimeoutError:
                    continue

                if file_output_queue:
                    file_output_queue.put({ 'timestamp': timestamp, 'spectrum': spectrum })

                if config.stdout:
                    print(format_spectrum(timestamp - start_time, spectrum))
                elif file_output_queue is None:
                    break

                continue

            try:
                range_value, signal_strength = sensor.get_single_target_list()
                timestamp = time()
            except TimeoutError:
                # Ignore timeouts and try again
                continue

            if range_value is None:
                range_value = 0.0

            if signal_strength is None:
                signal_strength = 0.0

            if file_output_queue:
                file_output_queue.put({ 'timestamp': timestamp, 'range': range_value, 'signal_strength': signal_strength })

            if config.stdout:
                print(format_value(timestamp - start_time, range_value, signal_strength))
            elif file_output_queue is None:
                print(f'Current range:    {range_value:7.4f} m')
                print(f'Current signal:   {signal_strength:7.2f} dB')
                break

        if file_output_queue and file_output_process:
            file_output_queue.put(None)
            file_output_process.join(1.0)
            if file_output_process.exitcode is None:
                file_output_process.kill()

        sensor.stop_acquisition()


if __name__ == '__main__':
    main()
