# pylint: disable=invalid-name
from enum import IntEnum
from math import ceil
from time import sleep
from typing import Tuple
import serial

# Transmit frame with variable length
# SD2  LE  LEr  SD2  DA  SA  FC  PDU  FCS  ED

# Transmit frame with fixed length
# SD2  DA  SA  FC  PDU  FCS  ED

SD2 = 0x68  # Start delimiter for frame with variable length used for sensor communicatio
SD3 = 0xa2  # Start delimiter for frame with fixed length used for iSYS-6030 answers
# LE, LEr    # Length of data for FC (DA, SA & FC)
# DA         # Destination/sensor address (default 100 = 0x64)
# Broadcast address = 0x0, master device adress = 0x1
# SA         # Source address
# FC         # Function code (incl. sub function code)
# PDU        # Protocol data unit (optional)
# FCS        # Frame checksum (addition of bytes from byte 4 to PDU, overflows are ignored)
ED = 0x16  # End delimiter

# Frames must be sent as a continous stream of data (max delay between bytes is about 10 ms)
# All accepted frames are acknowledged with a frame including the function code and possible requested data
# Incorrect or destroyed frames are ignored without replaying
# A bootloader message is sent at sensor power up (and after a reset_device command)
# Application firmware starts measuring after abour 225 ms


class MeasurementMode(IntEnum):
    """Single target mode 50 Hz"""
    SINGLE_TARGET_MODE = 0
    """Multi target mode (up to 10 targets) 10 Hz"""
    MULTI_TARGET_MODE_10HZ = 1
    """Long integration mode (improved SNR) 4 Hz"""
    LONG_INTEGRATION_MODE = 2
    """Multi target mpde (up to 15 targets) 25 Hz"""
    MULTI_TARGET_MODE_25HZ = 3


class SingleTargetFilterType(IntEnum):
    HIGHEST_AMPLITUDE = 0
    MEAN = 1
    MEDIAN = 2
    MIN = 3
    MAX = 4


class SingleTargetFilterSignal(IntEnum):
    OFF = 0
    RANGE_RADIAL = 2


class SensorError(Exception):
    pass


class Sensor:

    def __init__(self, port, *, baudrate=115200, sensor_address=100) -> None:
        self._serial = serial.Serial(
            port, baudrate, timeout=1, inter_byte_timeout=1)
        self._sensor_address = sensor_address

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.close()

    def close(self):
        self._serial.close()

# --------------- iSYS API ----------------------------
    def reset_device(self) -> None:
        """ Soft reset of sensor """
        LE = 0x05
        SA = 0x01
        FC = [0xbc, 0x00, 0x01]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        sleep(1)
        ack, _ = self._read_response()
        # Check and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

    def get_device_name(self) -> Tuple[str, int]:
        """ Read decvice name and serial number """
        LE = 0x03
        SA = 0x01
        FC = [0xd0]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        # Check and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        if s:
            device_name, device_serial_number = s.split(b'_')
        else:
            raise SensorError('Invalid response')
        return device_name.decode('ascii'), int(device_serial_number[:-1])

    def start_acquisition(self) -> None:
        self._acquisition(1)

    def stop_acquisition(self) -> None:
        self._acquisition(0)

    def get_temperature(self):
        """ Read sensor temperature """
        LE = 0x05
        SA = 0x01
        FC = [0xd1, 0x01, 0x09]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        return ((s[0] << 8) + s[1])/100

    def get_address(self) -> int:
        """ Read sensor address """
        LE = 0x05
        SA = 0x01
        FC = [0xd2, 0x00, 0x01]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        return int((s[0] << 8) + s[1])

    def set_address(self, address: int):
        """ Write sensor address """
        # Check input
        if address < 2 or address > 0xff:
            raise ValueError('Invalid sensor address')
        LE = 0x07
        SA = 0x01
        FC = [0xd3, 0x00, 0x01, 0x00, address]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, _ = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

    def get_mode(self) -> MeasurementMode:
        """ Read measurement mode """
        #   Mode
        #   0 - Single target mode                      50 Hz
        #   1 - Multi target mode (up to 10 targets)    10 Hz
        #   2 - Long integration mode (improved SNR)     4 Hz
        #   3 - Multi target mpde (up to 15 targets)    25 Hz

        LE = 0x05
        SA = 0x01
        FC = [0xd2, 0x00, 0x10]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        return MeasurementMode(int((s[0] << 8) + s[1]))

    def set_mode(self, mode: MeasurementMode):
        """ Write sensor measurement mode """
        #   Mode
        #   0 - Single target mode                      50 Hz
        #   1 - Multi target mode (up to 10 targets)    10 Hz
        #   2 - Long integration mode (improved SNR)     4 Hz
        #   3 - Multi target mpde (up to 15 targets)    25 Hz

        # Check input
        if mode not in set(item.value for item in MeasurementMode):
            raise ValueError(f'Invalid measurement mode: {mode}')
        LE = 0x07
        SA = 0x01
        FC = [0xd3, 0x00, 0x10, 0x00, mode]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, _ = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

    def get_threshold(self) -> float:
        """ Read sensor threshold sensitivity """
        LE = 0x05
        SA = 0x01
        FC = [0xd2, 0x00, 0x16]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        return int((s[0] << 8) + s[1])/10

    def set_threshold(self, threshold: float):
        """ Write sensor threshold in dB """
        LE = 0x07
        SA = 0x01
        FC = [0xd3, 0x00, 0x16, 0x00, int(threshold*10)]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, _ = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

    def get_minimum_range(self) -> float:
        return self._get_app_settings('min_range')

    def get_maximum_range(self) -> float:
        return self._get_app_settings('max_range')

    def get_minimum_signal(self) -> float:
        return self._get_app_settings('min_signal')

    def get_filter_type(self) -> int:
        return SingleTargetFilterType(self._get_app_settings('filter_type'))

    def get_filter_signal(self) -> int:
        return SingleTargetFilterSignal(self._get_app_settings('filter_signal'))

    def get_firmware_version(self) -> str:
        return self._get_cal_settings('firmware')

    def get_hardware_version(self) -> str:
        return self._get_cal_settings('hardware')

    def get_product_code(self) -> str:
        return self._get_cal_settings('product_code')

    def get_bootloader_version(self) -> str:
        return self._get_cal_settings('bootloader')

    def set_minimum_range(self, minimum_range) -> None:
        self._set_app_settings('min_range', minimum_range)

    def set_maximum_range(self, maximum_range) -> None:
        self._set_app_settings('max_range', maximum_range)

    def set_minimum_signal(self, minimum_signal) -> None:
        self._set_app_settings('min_signal', minimum_signal)

    def set_maximum_signal(self, maximum_signal) -> None:
        self._set_app_settings('max_signal', maximum_signal)

    def set_filter_type(self, filter_type: SingleTargetFilterType) -> None:
        self._set_app_settings('filter_type', filter_type)

    def set_filter_signal(self, filter_signal: SingleTargetFilterSignal) -> None:
        self._set_app_settings('filter_signal', filter_signal)

    def get_single_target_list(self) -> Tuple[float, float]:
        """ Read single target list """
        # Set timeout to max frame length to receive (in sec)
        self._serial.timeout = 0.02
        LE = 0x05
        SA = 0x01
        FC = [0xd9, 0x01, 0x00]
        # Send frame, wait min 5 ms
        sleep(0.005)
        self._serial.reset_output_buffer()
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        self._serial.timeout = 1
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

        output_no = s[0]
        no_of_targets = s[1]
        if output_no != 1 or no_of_targets != 1:
            return None, None

        signal_target = ((s[2] << 8) + s[3])/100
        range_target = ((s[4] << 24) + (s[5] << 16) +
                        (s[6] << 8) + s[7])/1000000
        return range_target, signal_target

    def get_range_spectrum(self):
        """ Read range spectrum - single target list """
        # Set timeout according to packet size
        self._serial.timeout = 0.5 # Large packet
        LE = 0x03
        SA = 0x01
        FC = [0xdd]
        # Send frame
        self._serial.reset_output_buffer()
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, fft_length, s = self._read_legacy_response()
        self._serial.timeout = 1
        # Check address & code and return response
        if not self._check_legacy_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        # Get FFT data and threshold
        data = [0]*fft_length
        threshold = [0]*fft_length
        index = 0
        for i in range(0,fft_length*4,4):
            data[index] = ((s[i] << 8) + s[i+1]) / 10
            threshold[index] = ((s[i+2] << 8) + s[i+3]) / 10
            index += 1
        no_detections = s[-1]
        return data, threshold, no_detections

    def set_factory_settings(self) -> None:
        """ Set factory settings (in flash) """
        LE = 0x04
        SA = 0x01
        FC = [0xdf, 0x01]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, _ = self._read_response()
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

    def save_settings(self) -> None:
        """ Save sensor/application settings to flash """
        LE = 0x04
        SA = 0x01
        FC = [0xdf, 0x04]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, _ = self._read_response()
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

# ------------------------------------------------------------------------------
# Local functions
    def _get_app_settings(self, parameter):
        """ Read application setting """
        # Sub fuction code
        # 0x0108 - Min range in m
        # 0x0109 - Max range in m
        # 0x010A - Min signal strength in dB
        # 0x010B - Max signal strength in dB
        # 0x0115 - Type of signal target filter
        #          0 = highest amplitude
        #          1 = mean
        #          2 = median
        #          3 = min (closest)
        #          4 = max (furthest)
        # 0x0116 - Signal for singel target filter (0 = off, 2 = range radial)
        # 0x070C - Type of signal target filter (4 x uint8-t or 1 x float32_t)

        LE = 0x05
        SA = 0x01

        factor = 0.1
        if parameter == 'min_range':
            FC = [0xd4, 0x01, 0x08]
        if parameter == 'max_range':
            FC = [0xd4, 0x01, 0x09]
        if parameter == 'min_signal':
            FC = [0xd4, 0x01, 0x0a]
        if parameter == 'max_signal':
            FC = [0xd4, 0x01, 0x0b]
        if parameter == 'filter_type':
            FC = [0xd4, 0x01, 0x15]
            factor = 1
        if parameter == 'filter_signal':
            FC = [0xd4, 0x01, 0x16]
            factor = 1
        # if parameter == 'signal_type':
        #     FC = [0xd4, 0x07, 0x0c]
        #     factor = 1

        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        # Check address & code
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        # Return response
        return ((s[0] << 8) + s[1])*factor

    def _set_app_settings(self, parameter, val):
        """ Write application setting """
        # Sub fuction code
        # 0x0108 - Min range in 0.1 m (0 - 40/20 m for mode 0/2)
        # 0x0109 - Max range in 0.1 m (0 - 40/20 m for mode 0/2)
        # 0x010A - Min signal strength in 0.1 dB (20 - 80 dB reccomended)
        # 0x010B - Max signal strength in 0.1 dB (20 - 80 dB reccomended)
        # 0x0115 - Type of signal target filter
        #          0 = highest amplitude
        #          1 = mean
        #          2 = median
        #          3 = min (closest)
        #          4 = max (furthest)
        # 0x0116 - Signal for singel target filter (0 = off, 2 = range radial)
        # 0x070C - Type of signal target filter (4xuint8-t or 1xfloat32_t)

        LE = 0x07
        SA = 0x01

        if parameter == 'min_range':
            data = int(val*10)
            FC = [0xd5, 0x01, 0x08, data >> 8, data & 0xff]
        if parameter == 'max_range':
            data = ceil(val*10)
            FC = [0xd5, 0x01, 0x09, data >> 8, data & 0xff]
        if parameter == 'min_signal':
            data = int(val*10)
            FC = [0xd5, 0x01, 0x0a, data >> 8, data & 0xff]
        if parameter == 'max_signal':
            data = ceil(val*10)
            FC = [0xd5, 0x01, 0x0b, data >> 8, data & 0xff]
        if parameter == 'filter_type':
            data = val
            FC = [0xd5, 0x01, 0x15, data >> 8, data & 0xff]
        if parameter == 'filter_signal':
            data = val
            FC = [0xd5, 0x01, 0x16, data >> 8, data & 0xff]
        # if parameter == 'signal_type':
        #     data = val*1
        #     FC = [0xd5, 0x07, 0x0c, data>>8, data&0xff]

        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, _ = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

    def _get_cal_settings(self, parameter):
        """ Read calibration setting """
        # Sub fuction code
        # 0x0101 - Firmaware version
        # 0x0102 - DSP hardware version
        # 0x0104 - Prodcut code
        # 0x0220 - Bootloader version

        LE = 0x05
        SA = 0x01

        if parameter == 'firmware':
            FC = [0xd6, 0x01, 0x01]
        if parameter == 'hardware':
            FC = [0xd6, 0x01, 0x02]
        if parameter == 'product_code':
            FC = [0xd6, 0x01, 0x04]
        if parameter == 'bootloader':
            FC = [0xd6, 0x02, 0x20]

        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, s = self._read_response()
        # Check address & code
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')
        # Return response
        if parameter == 'product_code':
            return (s[0] << 8) + s[1]
        major_ver = (s[0] << 8) + s[1]
        minor_ver = (s[4] << 8) + s[5]
        ver = f'{major_ver}.{minor_ver:03d}'
        return ver

    def _acquisition(self, start):
        """ Start/stop sensor measurement (1/0) """
        LE = 0x05
        SA = 0x01
        FC = [0xd1, 0x00, 0x00] if start else [0xd1, 0x00, 0x01]
        # Send frame
        self._serial.write(self._write_frame(LE, SA, FC))
        # Get acknowledge and response
        ack, _ = self._read_response()
        # Check address & code and return response
        if not self._check_ack_frame(ack, FC[0]):
            raise SensorError('Invalid response')

    def _write_frame(self, length, sub_address, code):
        array = bytearray(
            [SD2, length, length, SD2, self._sensor_address, sub_address])
        for i in code:
            array.extend(bytes([i]))
        # code (FC) can be 1 or 3 bytes (incl. sub-code)
        fcs = self._frame_checksum(array[4:4+2+len(code)])
        array.extend(bytes([fcs, ED]))
        return array

    def _read_response(self):
        self._serial.reset_input_buffer()
        # Read acknowledge packet (7 bits)
        ack = self._serial.read(7)
        if len(ack) < 7:
            raise TimeoutError()
        tmp = self._serial.read(ack[1]-1)  # read remaining part of frame
        data = tmp[:-2]  # skip FCS & ED
        return ack, data

    def _read_legacy_response(self):
        self._serial.reset_input_buffer()
        # Read acknowledge packet (7 bits)
        ack = self._serial.read(4)
        if len(ack) < 4:
            raise TimeoutError()
        length = self._serial.read(1)
        fft_length = (length[0] & 0x7f) * 16
        tmp = self._serial.read(fft_length*4+3)  # read remaining part of frame
        fft_data = tmp[:-2]  # skip FCS & ED
        return ack, fft_length, fft_data

    def _frame_checksum(self, array):
        """ Calculate checksum by adding all bytes in array - ingoring overflows """
        return int(hex((sum(array)) & 0xff), 16)

    def _check_ack_frame(self, ack, code):
        """ Check sensor address and function code (FC[0]) """
        if ack[5] == self._sensor_address and ack[6] == code:
            return True
        return False

    def _check_legacy_ack_frame(self, ack, code):
        """ Check sensor address and function code (FC[0]) """
        if ack[2] == self._sensor_address and ack[3] == code:
            return True
        return False
